import numpy as np
from mpi4py import MPI
from .file_base import FileBase

__all__ = ('HDF5File',)

comm = MPI.COMM_WORLD

class HDF5File(FileBase):
    """Class for reading/writing data to HDF5 format

    Parameters
    ----------
        h5name : str
            Name of hdf5 file to be created.
        T : PFFT
            Instance of a :class:`PFFT` class. Must be the same as the space
            used for storing with :class:`HDF5Writer.write`.
        domain : Sequence, optional
            The spatial domain. Sequence of either

                - 2-tuples, where each 2-tuple contains the (origin, length)
                  of each dimension, e.g., (0, 2*pi).
                - Arrays of coordinates, e.g., np.linspace(0, 2*pi, N). One
                  array per dimension.
        mode : str, optional
            ``r`` or ``w`` for read or write. Default is ``r``.
    """
    def __init__(self, h5name, T, domain=None, mode='r', **kw):
        FileBase.__init__(self, T, domain=domain, **kw)
        import h5py
        self.f = h5py.File(h5name, mode, driver="mpio", comm=comm)
        if mode == 'w':
            if isinstance(self.domain[0], np.ndarray):
                self.f.create_group("mesh")
            else:
                self.f.create_group("domain")
            for i in range(T.ndim()):
                d = self.domain[i]
                if isinstance(d, np.ndarray):
                    self.f["mesh"].create_dataset("x{}".format(i), data=np.squeeze(d))
                else:
                    self.f["domain"].create_dataset("x{}".format(i), data=np.array([d[0], d[1]]))
            self.f.attrs.create("ndim", T.ndim())
            self.f.attrs.create("shape", T.shape(False))

    @staticmethod
    def backend():
        return 'hdf5'

    def write(self, step, fields, **kw):
        """Write snapshot ``step`` of ``fields`` to HDF5 file

        Parameters
        ----------
        step : int
            Index of snapshot.
        fields : dict
            The fields to be dumped to file. (key, value) pairs are group name
            and either arrays or 2-tuples, respectively. The arrays are complete
            arrays to be stored, whereas 2-tuples are arrays with associated
            *global* slices.
        forward_output : bool, optional
            Whether fields to be stored are shaped as the output of a
            forward transform or not. Default is False.

        Example
        -------
        >>> from mpi4py import MPI
        >>> from mpi4py_fft import PFFT, HDF5File, Function
        >>> comm = MPI.COMM_WORLD
        >>> T = PFFT(comm, (15, 16, 17))
        >>> u = Function(T, forward_output=False, val=1)
        >>> v = Function(T, forward_output=False, val=2)
        >>> f = HDF5File('h5filename.h5', T)
        >>> f.write(0, {'u': [u, (u, [slice(None), 4, slice(None)])],
        ...             'v': [v, (v, [slice(None), 5, 5])]})
        >>> f.write(1, {'u': [u, (u, [slice(None), 4, slice(None)])],
        ...             'v': [v, (v, [slice(None), 5, 5])]})

        This stores data within two main groups ``u`` and ``v``. The HDF5 file
        will in the end contain groups::

            /u/3D/{0, 1}
            /u/2D/slice_4_slice/{0, 1}
            /v/3D/{0, 1}
            /v/1D/slice_5_5/{0, 1}

        Note
        ----
        The list of slices used in storing only parts of the arrays are views
        of the *global* arrays.

        """
        forward_output = kw.get('forward_output', False)
        FileBase.write(self, step, fields, forward_output=forward_output)

    def read(self, u, name, **kw):
        """Read into array ``u``

        Parameters
        ----------
        u : array
            The array to read into.
        name : str
            Name of array to be read.
        forward_output : bool, optional
            Whether the array to be read is the output of a forward transform
            or not. Default is False.
        step : int, optional
            Index of field to be read. Default is 0.
        """
        forward_output = kw.get('forward_output', False)
        step = kw.get('step', 0)
        s = self.T.local_slice(forward_output)
        ndim = self.T.ndim()
        dset = "/".join((name, "{}D".format(ndim), str(step)))
        u[:] = self.f[dset][tuple(s)]

    def _write_slice_step(self, name, step, slices, field, **kw):
        forward_output = kw.get('forward_output', False)
        slices = list(slices)
        ndims = slices.count(slice(None))
        slname = self._get_slice_name(slices)
        s = self.T.local_slice(forward_output)
        slices, inside = self._get_local_slices(slices, s)
        sp = np.nonzero([isinstance(x, slice) for x in slices])[0]
        sf = tuple(np.take(s, sp))
        sl = tuple(slices)
        group = "/".join((name, "{}D".format(ndims), slname))
        if group not in self.f:
            self.f.create_group(group)
        N = self.T.shape(forward_output)
        self.f[group].create_dataset(str(step), shape=np.take(N, sp), dtype=field.dtype)
        if inside == 1:
            self.f["/".join((group, str(step)))][sf] = field[sl]

    def _write_group(self, name, u, step, **kw):
        forward_output = kw.get('forward_output', False)
        s = tuple(self.T.local_slice(forward_output))
        group = "/".join((name, "{}D".format(self.T.ndim())))
        if group not in self.f:
            self.f.create_group(group)
        self.f[group].create_dataset(str(step), shape=self.T.shape(forward_output), dtype=u.dtype)
        self.f["/".join((group, str(step)))][s] = u



import numpy as np
from mpi4py import MPI
from .file_base import FileBase

# https://github.com/Unidata/netcdf4-python/blob/master/examples/mpi_example.py
# Note. Not using groups because Visit does not understand it

__all__ = ('NCFile',)

comm = MPI.COMM_WORLD

class NCFile(FileBase):
    """Class for writing data to NetCDF4 format

    Parameters
    ----------
        ncname : str
            Name of netcdf file to be created
        T : PFFT
            Instance of a :class:`.PFFT` class. Must be the same as the space
            used for storing with :class:`NCWriter.write`
        domain : Sequence
            The spatial domain. Sequence of either

                - 2-tuples, where each 2-tuple contains the (origin, length)
                  of each dimension, e.g., (0, 2*pi).
                - Arrays of coordinates, e.g., np.linspace(0, 2*pi, N). One
                  array per dimension.
        clobber : bool, optional
        mode : str
            ``r`` or ``w`` for read or write. Default is 'r'.
    Note
    ----
    Each class instance creates one unique NetCDF4-file, with one step-counter.
    It is possible to store multiple fields in each file, but all snapshots of
    the fields must be taken at the same time. If you want one field stored
    every 10th timestep and another every 20th timestep, then use two different
    class instances and as such two NetCDF4-files.
    """
    def __init__(self, ncname, T, domain=None, clobber=True, mode='r', **kw):
        FileBase.__init__(self, T, domain=domain, **kw)
        from netCDF4 import Dataset
        self.f = Dataset(ncname, mode=mode, clobber=clobber, parallel=True, comm=comm, **kw)
        self.N = N = T.shape(False)[-T.ndim():]
        dtype = self.T.dtype(False)
        assert dtype.char in 'fdg'
        self._dtype = dtype

        if mode == 'w':
            self.f.createDimension('time', None)
            self.dims = ['time']
            self.nc_t = self.f.createVariable('time', self._dtype, ('time'))
            self.nc_t.set_collective(True)

            d = list(self.domain)
            if not isinstance(self.domain[0], np.ndarray):
                assert len(self.domain[0]) == 2
                for i in range(T.ndim()):
                    d[i] = np.arange(N[i], dtype=np.float)*2*np.pi/N[i]
            else:
                for i in range(T.ndim()):
                    d[i] = np.squeeze(d[i])
            self.domain = d
            for i in range(T.ndim()):
                xyz = 'xyzrst'[i]
                self.f.createDimension(xyz, N[i])
                self.dims.append(xyz)
                nc_xyz = self.f.createVariable(xyz, self._dtype, (xyz))
                nc_xyz[:] = d[i]
            self.f.setncatts({"ndim": T.ndim(), "shape": T.shape(False)})
            self.handles = dict()
            self.f.sync()

    @staticmethod
    def backend():
        return 'netcdf4'

    def write(self, step, fields, **kw):
        """Write snapshot step of ``fields`` to NetCDF4 file

        Parameters
        ----------
        step : int
            Index of snapshot
        fields : dict
            The fields to be dumped to file. (key, value) pairs are group name
            and either arrays or 2-tuples, respectively. The arrays are complete
            arrays to be stored, whereas 2-tuples are arrays with associated
            *global* slices.
        """
        it = self.nc_t.size
        self.nc_t[it] = step
        FileBase.write(self, it, fields)

    def read(self, u, name, **kw):
        """Read into array ``u``

        Parameters
        ----------
        u : array
            The array to read into.
        name : str
            Name of array to be read.
        step : int, optional
            Index of field to be read. Default is 0.
        """
        step = kw.get('step', 0)
        s = self.T.local_slice(False)
        s = [step] + s
        u[:] = self.f[name][tuple(s)]

    def _write_slice_step(self, name, step, slices, field, **kw):
        slices = list(slices)
        slname = self._get_slice_name(slices)
        s = self.T.local_slice(False)

        slices, inside = self._get_local_slices(slices, s)
        sp = np.nonzero([isinstance(x, slice) for x in slices])[0]
        sf = np.take(s, sp)
        sdims = ['time'] + list(np.take(self.dims, np.array(sp)+1))
        fname = "_".join((name, slname))
        if fname not in self.handles:
            self.handles[fname] = self.f.createVariable(fname, self._dtype, sdims)
            self.handles[fname].set_collective(True)

        self.handles[fname][step] = 0 # collectively create dataset
        self.handles[fname].set_collective(False)
        sf = tuple([step] + list(sf))
        sl = tuple(slices)
        if inside:
            self.handles[fname][sf] = field[sl]
        self.handles[fname].set_collective(True)
        self.f.sync()

    def _write_group(self, name, u, step, **kw):
        s = self.T.local_slice(False)
        if name not in self.handles:
            self.handles[name] = self.f.createVariable(name, self._dtype, self.dims)
            self.handles[name].set_collective(True)
        s = tuple([step] + s)
        self.handles[name][s] = u
        self.f.sync()



#!/usr/bin/env python

import os
import sys
import re
import shutil
from distutils import ccompiler
from setuptools import setup
from setuptools.extension import Extension
from numpy import get_include

cwd = os.path.abspath(os.path.dirname(__file__))
fftwdir = os.path.join(cwd, 'mpi4py_fft', 'fftw')
prec_map = {'float': 'f', 'double': '', 'long double': 'l'}

def get_library_dirs():
    lib_dirs = [os.path.join(sys.prefix, 'lib')]
    for f in ('FFTW_ROOT', 'FFTW_DIR'):
        if f in os.environ:
            lib_dirs.append(os.path.join(os.environ[f], 'lib'))
    return lib_dirs

def get_include_dirs():
    inc_dirs = [get_include(), os.path.join(sys.prefix, 'include')]
    for f in ('FFTW_ROOT', 'FFTW_DIR'):
        if f in os.environ:
            inc_dirs.append(os.path.join(os.environ[f], 'include'))
    return inc_dirs

def get_fftw_libs():
    """Return FFTW libraries"""
    compiler = ccompiler.new_compiler()
    library_dirs = get_library_dirs()
    libs = {}
    for d in ('float', 'double', 'long double'):
        lib = 'fftw3'+prec_map[d]
        if compiler.find_library_file(library_dirs, lib):
            libs[d] = [lib]
            tlib = '_'.join((lib, 'threads'))
            if compiler.find_library_file(library_dirs, tlib):
                libs[d].append(tlib)
            if sys.platform in ('unix', 'darwin'):
                libs[d].append('m')
    assert len(libs) > 0, "No FFTW libraries found in {} {}".format(library_dirs, sys.prefix)
    return libs

def generate_extensions(fftwlibs):
    """Generate files with float and long double"""
    for d in fftwlibs:
        if d == 'double':
            continue
        p = 'fftw'+prec_map[d]+'_'
        for fl in ('fftw_planxfftn.h', 'fftw_planxfftn.c', 'fftw_xfftn.pyx', 'fftw_xfftn.pxd'):
            fp = fl.replace('fftw_', p)
            shutil.copy(os.path.join(fftwdir, fl), os.path.join(fftwdir, fp))
            sedcmd = "sed -i ''" if sys.platform == 'darwin' else "sed -i''"
            os.system(sedcmd + " 's/fftw_/{0}/g' {1}".format(p, os.path.join(fftwdir, fp)))
            os.system(sedcmd + " 's/double/{0}/g' {1}".format(d, os.path.join(fftwdir, fp)))

def get_extensions(fftwlibs):
    """Return list of extension modules"""
    include_dirs = get_include_dirs()
    library_dirs = get_library_dirs()
    ext = [Extension("mpi4py_fft.fftw.utilities",
                     sources=[os.path.join(fftwdir, "utilities.pyx")],
                     include_dirs=include_dirs)]

    for d, libs in fftwlibs.items():
        p = 'fftw'+prec_map[d]+'_'
        ext.append(Extension("mpi4py_fft.fftw.{}xfftn".format(p),
                             sources=[os.path.join(fftwdir, "{}xfftn.pyx".format(p)),
                                      os.path.join(fftwdir, "{}planxfftn.c".format(p))],
                             #define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
                             libraries=libs,
                             include_dirs=include_dirs,
                             library_dirs=library_dirs))
    return ext

def version():
    srcdir = os.path.join(cwd, 'mpi4py_fft')
    with open(os.path.join(srcdir, '__init__.py')) as f:
        m = re.search(r"__version__\s*=\s*'(.*)'", f.read())
        return m.groups()[0]

with open("README.rst", "r") as fh:
    long_description = fh.read()

if __name__ == '__main__':
    fftw_libs = get_fftw_libs()
    generate_extensions(fftw_libs)
    setup(name="mpi4py-fft",
          version=version(),
          description="mpi4py-fft -- FFT with MPI",
          long_description=long_description,
          author="Lisandro Dalcin and Mikael Mortensen",
          url='https://bitbucket.org/mpi4py/mpi4py-fft',
          packages=["mpi4py_fft",
                    "mpi4py_fft.fftw",
                    "mpi4py_fft.utilities"],
          package_dir={"mpi4py_fft": "mpi4py_fft"},
          classifiers=[
              'Development Status :: 4 - Beta',
              'Environment :: Console',
              'Intended Audience :: Developers',
              'Intended Audience :: Science/Research',
              'Programming Language :: Python',
              'Programming Language :: Python :: 2',
              'Programming Language :: Python :: 3',
              'License :: OSI Approved :: BSD License',
              'Topic :: Scientific/Engineering :: Mathematics',
              'Topic :: Software Development :: Libraries :: Python Modules',
              ],
          ext_modules=get_extensions(fftw_libs),
          install_requires=["mpi4py", "numpy"],
          setup_requires=["setuptools>=18.0", "cython>=0.25"]
          )



"""
This is the **mpi4py-fft** package

What is **mpi4py-fft**?
=======================

The Python package **mpi4py-fft** is a tool primarily for working with Fast
Fourier Transforms (FFTs) of (large) multidimensional arrays. There is really
no limit as to how large the arrays can be, just as long as there is sufficient
computing powers available. Also, there are no limits as to how transforms can
be configured. Just about any combination of transforms from the FFTW library
is supported. Furthermore, mpi4py-fft can also be used simply to perform global
redistributions (distribute and communicate) of large arrays with MPI, without
any transforms at all.

For more information, see `documentation <https://mpi4py-fft.readthedocs.io>`_.

"""
__version__ = '1.1.0'
__author__ = 'Lisandro Dalcin and Mikael Mortensen'

from .mpifft import PFFT, Function
from . import fftw
from .utilities import HDF5File, NCFile, generate_xdmf
